// generated by verigenSV.pl on Fri Apr 30 17:16:31 2021
// input file: ras.cfg

// compiler directives
/* verilator lint_off CASEINCOMPLETE */
/* verilator lint_off DECLFILENAME */
/* verilator lint_off UNUSED */
/* verilator lint_off STMTDLY */
/* verilator lint_off UNOPTFLAT */

module RAS (
	input logic clk,
	input logic res_n,
	input logic lock,
	input logic unlock,
	input logic [KEY_WIDTH-1:0] key,
	input logic in_addr_val,
	input logic [RAS_WIDTH-1:0] in_addr,
	input logic rd,
	output logic out_addr_val, 
	output logic [RAS_WIDTH-1:0] out_addr, 
	output logic locked, 
	output logic ras_write_active, 
	output logic empty, 
	output logic full );

  //
 // parameter constants 
//---------------------------
parameter int RAS_SIZE = 16;
parameter int RAS_WIDTH = 48;
parameter int KEY_WIDTH = 32;
parameter logic [KEY_WIDTH-1:0] KEY = 32'hDEADBEEF;
parameter logic [RAS_WIDTH-1:0] ZERO_ADDR = {RAS_WIDTH{1'b0}}; ;


  //
 // state encoding typedef 
//--------------------------------
typedef enum logic [0:0] {
	LOCKED	 = 1'b1,
	UNLOCKED	 = 1'b0
} states_t;


// internal state variables
states_t STATE, NEXT_STATE;


  //
 // memory instantiations 
//---------------------------
logic RAS_STACK_status;
logic RAS_STACK_clk; 
logic RAS_STACK_res_n; 
logic RAS_STACK_wr;
logic [RAS_WIDTH-1:0] RAS_STACK_wr_data;
logic RAS_STACK_rd;
logic [RAS_WIDTH-1:0] RAS_STACK_rd_data;
logic RAS_STACK_empty;
logic RAS_STACK_full;
assign RAS_STACK_clk = clk;
assign RAS_STACK_res_n = res_n;
STACK #(RAS_WIDTH,RAS_SIZE) RAS_STACK(
	.clk(RAS_STACK_clk), .res_n(RAS_STACK_res_n), 
	.wr(RAS_STACK_wr), .wr_data(RAS_STACK_wr_data[RAS_WIDTH-1:0]), 
	.rd(RAS_STACK_rd), .rd_data(RAS_STACK_rd_data[RAS_WIDTH-1:0]), 
	.empty(RAS_STACK_empty), .full(RAS_STACK_full) );


  //
 // implicit continuous assignments 
//----------------------------------
wire DO_IDLE; assign DO_IDLE = !lock & !unlock;
wire DO_LOCK ; assign DO_LOCK  =  lock & !unlock ;
wire DO_READ_STACK; assign DO_READ_STACK = (UNLOCKED==STATE) & rd;
wire DO_UNLOCK; assign DO_UNLOCK = !lock &  unlock & IS_KEY_MATCH ;
wire DO_WRITE_STACK; assign DO_WRITE_STACK = (UNLOCKED==STATE) & in_addr_val;
wire IS_KEY_MATCH; assign IS_KEY_MATCH = (KEY==key);
wire RAS_STACK_rd_data; assign RAS_STACK_rd_data = READ_RAS_STACK(DO_READ_STACK);


  //
 // initials for simulator
//---------------------------
initial begin
  STATE = LOCKED;
  RAS_STACK_wr = 1'b0;
  RAS_STACK_rd = 1'b0;
  RAS_STACK_wr_data[RAS_WIDTH-1:0] = {RAS_WIDTH{1'b0}};

  // dump output (Verilator)
  if ($test$plusargs("trace") != 0) begin
    $display("[%0t] Tracing to logs/vlt_dump.vcd...\n", $time);
    $dumpfile("logs/vlt_dump.vcd");
    $dumpvars();
  end
  $display("[%0t] Model running...\n", $time);


end


  //
 // state transfer
//---------------------------
always_ff @( posedge clk ) begin
  if( res_n )
    STATE <= NEXT_STATE;
  else
    STATE <= LOCKED;

  // condition for $finish (Verilator)
  if( lock==1'b1 && unlock==1'b1 && in_addr_val==1'b1 && res_n==1'b0 ) begin
    $write("*-* All Finished *-*\n");
    $finish;
  end   

end
// AssertionExample:
//always_ff @ (posedge clk) begin
//   AssertionExample: assert (rd==1'b1 ^ wr==1'b1);
//end

// coverage analysis (Verilator)
// cover property ( @(posedge clk) rd_q0==1'b1 && rd_q1==1'b1 && res_n==1'b0 );

always_comb begin
  unique case ( STATE )
  LOCKED : begin
	// ATTENTION: secure that conditions in parallel if-clauses are mutually exclusive to safeguard determinism!
	if ( !DO_UNLOCK ) begin
	  NEXT_STATE = LOCKED;
	  // MEMORY action ...
	  RAS_STACK_status = WRITE_RAS_STACK(DO_WRITE_STACK,in_addr);
	  // MEALY output action ...
	  ras_write_active=1'b0;
	  out_addr_val=1'b0;
	  out_addr=ZERO_ADDR;
	  empty=1'b0;
	  full=1'b0;
	  locked=STATE;
	end
	if ( DO_UNLOCK ) begin
	  NEXT_STATE = UNLOCKED;
	  // MEMORY action ...
	  RAS_STACK_status = WRITE_RAS_STACK(DO_WRITE_STACK,in_addr);
	  // MEALY output action ...
	  ras_write_active=RAS_STACK_status;
	  out_addr_val=DO_READ_STACK;
	  out_addr=RAS_STACK_rd_data;
	  empty=RAS_STACK_empty;
	  full=RAS_STACK_full;
	  locked=STATE;
	end
  end // LOCKED
  UNLOCKED : begin
	// ATTENTION: secure that conditions in parallel if-clauses are mutually exclusive to safeguard determinism!
	if ( DO_IDLE ) begin
	  NEXT_STATE = UNLOCKED;
	  // MEMORY action ...
	  RAS_STACK_status = WRITE_RAS_STACK(DO_WRITE_STACK,in_addr);
	  // MEALY output action ...
	  ras_write_active=RAS_STACK_status;
	  out_addr_val=DO_READ_STACK;
	  out_addr=RAS_STACK_rd_data;
	  empty=RAS_STACK_empty;
	  full=RAS_STACK_full;
	  locked=STATE;
	end
	if ( DO_LOCK ) begin
	  NEXT_STATE = LOCKED;
	  // MEMORY action ...
	  RAS_STACK_status = WRITE_RAS_STACK(DO_WRITE_STACK,in_addr);
	  // MEALY output action ...
	  ras_write_active=1'b0;
	  out_addr_val=1'b0;
	  out_addr=ZERO_ADDR;
	  empty=1'b0;
	  full=1'b0;
	  locked=STATE;
	end
  end // UNLOCKED
  endcase
end // always_comb


  //
 // memory functions
//---------------------------
function automatic [RAS_WIDTH-1:0] READ_RAS_STACK( read );
   RAS_STACK_rd = read;
   READ_RAS_STACK = RAS_STACK_rd_data;
endfunction

function automatic WRITE_RAS_STACK( write, [RAS_WIDTH-1:0] write_data);
   RAS_STACK_wr = write;
   RAS_STACK_wr_data = write_data;
   WRITE_RAS_STACK = write;
endfunction

endmodule


  //
 // memory
//---------------------------
module STACK ( clk, res_n, wr, wr_data, rd, rd_data, empty, full );

parameter data_width = 32;
parameter num_of_entries = 16;
parameter addr_width = $clog2(num_of_entries);

input  logic clk, res_n, rd, wr;
input  logic [data_width-1:0] wr_data;
output logic [data_width-1:0] rd_data;
output logic empty, full;
    
parameter ZERO_ADDR = {addr_width{1'b0}};
parameter ZERO_DATA = {data_width{1'b0}};

logic [data_width-1:0] MEMORY [num_of_entries-1:0];
logic [addr_width-1:0]   HEAD, HEAD_PLUS_1, HEAD_MINUS_1; 

integer i;

initial begin
    for ( i = 0; i < num_of_entries; i = i + 1 ) begin
        MEMORY[ i ] = ZERO_DATA;
    end
    HEAD = ZERO_ADDR;
end

// IN-CYCLE READ ACCESS
assign rd_data = ( rd && res_n ) ? MEMORY[HEAD] : ZERO_DATA;
assign empty = (HEAD=={addr_width{1'b0}});
assign full = (HEAD=={addr_width{1'b1}});
assign HEAD_PLUS_1 = HEAD + 1'b1;
assign HEAD_MINUS_1 = HEAD - 1'b1;

always @ ( negedge clk ) begin
  if( res_n ) begin
    if( rd & !wr &&  !empty ) HEAD <= HEAD_MINUS_1; // READ ACCESS
    if( wr & ( !full | rd )) begin // WRITE ACCESS
      if( rd ) begin
        MEMORY[HEAD] <= wr_data;
      end else begin
        MEMORY[HEAD_PLUS_1] <= wr_data;
        HEAD <= HEAD_PLUS_1;
      end
    end
  end else begin
    MEMORY[ZERO_ADDR] <= ZERO_DATA;
    HEAD <= ZERO_ADDR;
  end
end

endmodule
