// generated by verigenSV.pl on Fri Jun  4 15:36:19 2021
// input file: twe.cfg

// compiler directives
// /* verilator lint_off CASEINCOMPLETE */
/* verilator lint_off DECLFILENAME */
// /* verilator lint_off UNUSED */
// /* verilator lint_off STMTDLY */
/* verilator lint_off UNOPTFLAT */

`include "./RAS.sv";

module TWE (
	input logic clk,
	input logic res_n,
	input logic lock,
	input logic unlock,
	input logic [KEY_WIDTH-1:0] key,
	input logic [VA48_ADDR_WIDTH-1:0] virt_addr_in,
	input logic virt_addr_in_vld,
	input logic [VA48_PTE_PAGE_BASE_ADDR_WIDTH+VAXX_PTE_PAGE_STATUS_WIDTH-1:0] data_from_memory,
	input logic data_from_memory_vld,
	input logic [VAXX_MODE_WIDTH-1:0] mode,
	input logic [PA48_ADDR_WIDTH-1:0] pte_root_addr,
	output logic locked, 
	output logic [PA48_ADDR_WIDTH-1:0] phys_addr_out, 
	output logic phys_addr_out_vld, 
	output logic [PA48_ADDR_WIDTH-1:0] data_to_memory, 
	output logic data_to_memory_vld, 
	output logic twe_active );

  //
 // parameter constants 
//---------------------------
//----------------------
// key lock mechanism
//----------------------
parameter int KEY_WIDTH = 16;
parameter logic [KEY_WIDTH-1:0] KEY = 16'hAFFA;
//----------------------
// Virtuall addresses
//----------------------
parameter int VAXX_MODE_WIDTH = 2;
parameter int VA48_ADDR_WIDTH = 48;
parameter int VA48_INDEX_WIDTH = 12;
//----------------------
// Page Table Entries
//----------------------
parameter int VA48_PTE_ENTRY_WIDTH = 40;
parameter int VA48_PTE_PAGE_BASE_ADDR_WIDTH = 36;
parameter int VAXX_PTE_PAGE_STATUS_WIDTH = 4;
parameter int VA48_PTE_PAGE_OFFSET_WIDTH = 12;
parameter int VA48_PTE_OFFSET_SIZE = 9;
parameter int VA48_PTE_LEVEL0_HIGH = 47;
parameter int VA48_PTE_LEVEL0_LOW = 39;
parameter int VA48_PTE_LEVEL1_HIGH = 38;
parameter int VA48_PTE_LEVEL1_LOW = 30;
parameter int VA48_PTE_LEVEL2_HIGH = 29;
parameter int VA48_PTE_LEVEL2_LOW = 21;
parameter int VA48_PTE_LEVEL3_HIGH = 20;
parameter int VA48_PTE_LEVEL3_LOW = 12;
//----------------------
// Physical addresses
//----------------------
parameter int PA48_ADDR_WIDTH = 48;
parameter int PA48_INDEX_WIDTH = 12;
//----------------------
// Content Addressable Memory
//----------------------
// 48 bit virtual address mode
//----------------------
parameter int VA48_CAM_NUM_OF_ENTRIES = 16;
parameter int VA48_CAM_ENTRY_WIDTH = 72;
parameter int VA48_CAM_TAG_WIDTH = 36;
parameter logic [VA48_CAM_ENTRY_WIDTH-1:0] VA48_ZERO_PTE = {VA48_CAM_ENTRY_WIDTH{1'b0}};
//----------------------
// ZERO-DATA TO MEMORY
//----------------------
parameter logic [PA48_ADDR_WIDTH-1:0] ZERO_DATA_TO_MEMORY = {PA48_ADDR_WIDTH{1'b0}};
parameter logic [PA48_ADDR_WIDTH-1:0] ZERO_PHYS_ADDR_OUT = {PA48_ADDR_WIDTH{1'b0}};
parameter logic [VA48_CAM_ENTRY_WIDTH-1:0] ZERO_CAM_ENTRY = {VA48_CAM_ENTRY_WIDTH{1'b0}};


  //
 // state encoding typedef 
//--------------------------------
typedef enum logic [2:0] {
	LOCKED = 3'b111,
	IDLE = 3'b000,
	PTE0 = 3'b001,
	PTE1 = 3'b010,
	PTE2 = 3'b011,
	PTE3 = 3'b100,
	WRITE_TLB = 3'b101,
	DONE = 3'b110
} states_t;


// internal state variables
states_t STATE, NEXT_STATE;

  //
 // memory instantiations 
//---------------------------
uwire VA48_CAM_hit;
CAM #( VA48_CAM_ENTRY_WIDTH, VA48_CAM_NUM_OF_ENTRIES, VA48_CAM_TAG_WIDTH ) VA48_CAM(
	.clk(clk), .res_n(res_n), 
	.wr(VA48_CAM_wr), .wr_data(VA48_CAM_wr_data[VA48_CAM_ENTRY_WIDTH-1:0]), 
	.rd(VA48_CAM_rd), .rd_data(VA48_CAM_rd_data[VA48_CAM_ENTRY_WIDTH-1:0]), 
	.hit(VA48_CAM_hit) );

  //
 // include instantiations 
//---------------------------
uwire [32-1:0] ras_key_int;
uwire ras_in_addr_val_int;
uwire [48-1:0] ras_in_addr_int;
uwire ras_rd_int;
uwire ras_out_addr_val_int;
uwire [48-1:0] ras_out_addr_int;
uwire ras_locked_int;
uwire ras_write_active_int;
uwire ras_empty_int;
uwire ras_full_int ;
RAS #( 16, 48, 32, 32'h00000000 ) RAS16X48_K32( 
	clk, res_n, lock, unlock, ras_key_int[32-1:0], ras_in_addr_val_int, ras_in_addr_int[48-1:0], ras_rd_int, // inputs
	ras_out_addr_val_int, ras_out_addr_int[48-1:0], ras_locked_int, ras_write_active_int, ras_empty_int, ras_full_int  ); // outputs

   //
  // user-defined SystemVerilog support code 
 // (derived from <support> section) 
//----------------------------------
logic [VA48_CAM_ENTRY_WIDTH-1:0] data_from_tlb;
assign ras_key_int = 32'h00000000;
assign ras_in_addr_val_int = 1'b0;
assign ras_in_addr_int = '0;
assign ras_rd_int = 1'b0;

   //
  // implicit continuous assignments 
 // (derived from <expressions> section) 
//----------------------------------
assign locked = IS_LOCKED;
uwire DO_IDLE; assign DO_IDLE = !res_n || (IS_LOCKED && !unlock) || (!IS_LOCKED && !virt_addr_in_vld);
uwire DO_LOCK; assign DO_LOCK = lock & !unlock ;
uwire DO_UNLOCK; assign DO_UNLOCK = !lock &  unlock & IS_KEY_MATCH;
uwire IS_KEY_MATCH; assign IS_KEY_MATCH = (KEY==key);
uwire IS_LOCKED; assign IS_LOCKED = (STATE==LOCKED);
uwire IS_TLB_HIT; assign IS_TLB_HIT = (res_n && virt_addr_in_vld && VA48_CAM_hit);
uwire NEXT_LEVEL; assign NEXT_LEVEL = (res_n && virt_addr_in_vld && data_from_memory_vld && !VA48_CAM_hit);
uwire START_TWE; assign START_TWE = (res_n && virt_addr_in_vld);

  //
 // initials for simulator
//---------------------------
initial begin
  STATE = LOCKED;
  VA48_CAM_wr = 1'b0;
  VA48_CAM_rd = 1'b0;

  // dump output (Verilator)
  if ($test$plusargs("trace") != 0) begin
    $display("[%0t] Tracing to logs/vlt_dump.vcd...\n", $time);
    $dumpfile("logs/vlt_dump.vcd");
    $dumpvars();
  end
  $display("[%0t] Model running...\n", $time);

end

  //
 // state transfer
//---------------------------
always_ff @( posedge clk ) begin
  if( res_n )
    STATE <= NEXT_STATE;
  else
    STATE <= LOCKED;

  // $finish (Verilator)
  if( lock==1'b1 && unlock==1'b1 && res_n==1'b0 ) begin
    $write("*-* All Finished *-*\n");
    $finish;
  end   

end

// AssertionExample:
//always_ff @ (posedge clk) begin
//   AssertionExample: assert (rd==1'b1 ^ wr==1'b1);
//end

// coverage analysis (Verilator)
// cover property ( @(posedge clk) rd_q0==1'b1 && rd_q1==1'b1 && res_n==1'b0 );

always_comb begin

    //
   // pre-case (default) assignments 
  //----------------------------------
  data_from_tlb = ZERO_CAM_ENTRY;
  data_to_memory = ZERO_DATA_TO_MEMORY;
  data_to_memory_vld = 1'b0;
  phys_addr_out = ZERO_PHYS_ADDR_OUT;
  phys_addr_out_vld = 1'b0;
  twe_active = 1'b0;
  NOP_VA48_CAM();

  unique case ( STATE )
  LOCKED : begin
	if ( !DO_UNLOCK ) begin
	  NEXT_STATE = LOCKED;
	end
	if ( DO_UNLOCK ) begin
	  NEXT_STATE = IDLE;
	end
  end // LOCKED
  IDLE : begin
	if ( DO_IDLE ) begin
	  NEXT_STATE = IDLE;
	end
	if ( DO_LOCK ) begin
	  NEXT_STATE = LOCKED;
	end
	if ( START_TWE ) begin
	  NEXT_STATE = PTE0;
	  twe_active=1'b1;
	end
  end // IDLE
  PTE0 : begin
	if ( !NEXT_LEVEL ) begin
	  NEXT_STATE = PTE0;
	  data_from_tlb=POP_VA48_CAM(1'b1,{virt_addr_in[VA48_ADDR_WIDTH-1:VA48_INDEX_WIDTH],{VA48_PTE_PAGE_BASE_ADDR_WIDTH{1'b0}}});
	  twe_active=1'b1;
	  data_to_memory={pte_root_addr[PA48_ADDR_WIDTH-1:VA48_PTE_OFFSET_SIZE],virt_addr_in[VA48_PTE_LEVEL0_HIGH:VA48_PTE_LEVEL0_LOW]};
	  data_to_memory_vld=1'b1;
	end
	if ( IS_TLB_HIT ) begin
	  NEXT_STATE = DONE;
	  twe_active=1'b1;
	  phys_addr_out={data_from_tlb[VA48_PTE_PAGE_BASE_ADDR_WIDTH-1:0],virt_addr_in[VA48_INDEX_WIDTH-1:0]};
	  phys_addr_out_vld=1'b1;
	end
	if ( DO_IDLE ) begin
	  NEXT_STATE = IDLE;
	end
	if ( NEXT_LEVEL ) begin
	  NEXT_STATE = PTE1;
	  twe_active=1'b1;
	  data_to_memory={data_from_memory[VA48_PTE_PAGE_BASE_ADDR_WIDTH-1:0],virt_addr_in[VA48_PTE_LEVEL1_HIGH:VA48_PTE_LEVEL1_LOW],3'b000};
	end
  end // PTE0
  PTE1 : begin
	if ( !NEXT_LEVEL ) begin
	  NEXT_STATE = PTE1;
	  twe_active=1'b1;
	  data_to_memory={data_from_memory[VA48_PTE_PAGE_BASE_ADDR_WIDTH-1:0],virt_addr_in[VA48_PTE_LEVEL1_HIGH:VA48_PTE_LEVEL1_LOW],3'b000};
	  data_to_memory_vld=1'b1;
	end
	if ( NEXT_LEVEL ) begin
	  NEXT_STATE = PTE2;
	  twe_active=1'b1;
	  data_to_memory={data_from_memory[VA48_PTE_PAGE_BASE_ADDR_WIDTH-1:0],virt_addr_in[VA48_PTE_LEVEL2_HIGH:VA48_PTE_LEVEL2_LOW],3'b000};
	end
	if ( DO_IDLE ) begin
	  NEXT_STATE = IDLE;
	end
  end // PTE1
  PTE2 : begin
	if ( !NEXT_LEVEL ) begin
	  NEXT_STATE = PTE2;
	  twe_active=1'b1;
	  data_to_memory={data_from_memory[VA48_PTE_PAGE_BASE_ADDR_WIDTH-1:0],virt_addr_in[VA48_PTE_LEVEL2_HIGH:VA48_PTE_LEVEL2_LOW],3'b000};
	  data_to_memory_vld=1'b1;
	end
	if ( NEXT_LEVEL ) begin
	  NEXT_STATE = PTE3;
	  twe_active=1'b1;
	  data_to_memory={data_from_memory[VA48_PTE_PAGE_BASE_ADDR_WIDTH-1:0],virt_addr_in[VA48_PTE_LEVEL3_HIGH:VA48_PTE_LEVEL3_LOW],3'b000};
	end
	if ( DO_IDLE ) begin
	  NEXT_STATE = IDLE;
	end
  end // PTE2
  PTE3 : begin
	if ( !NEXT_LEVEL ) begin
	  NEXT_STATE = PTE3;
	  twe_active=1'b1;
	  data_to_memory={data_from_memory[VA48_PTE_PAGE_BASE_ADDR_WIDTH-1:0],virt_addr_in[VA48_PTE_LEVEL3_HIGH:VA48_PTE_LEVEL3_LOW],3'b000};
	  data_to_memory_vld=1'b1;
	end
	if ( NEXT_LEVEL ) begin
	  NEXT_STATE = WRITE_TLB;
	  twe_active=1'b1;
	end
	if ( DO_IDLE ) begin
	  NEXT_STATE = IDLE;
	end
  end // PTE3
  WRITE_TLB : begin
	if ( !DO_IDLE ) begin
	  NEXT_STATE = DONE;
	  PUSH_VA48_CAM(1'b1,{virt_addr_in[VA48_ADDR_WIDTH-1:VA48_INDEX_WIDTH],data_from_memory[VA48_PTE_PAGE_BASE_ADDR_WIDTH-1:0]});
	  twe_active=1'b1;
	  phys_addr_out={data_from_memory[VA48_PTE_PAGE_BASE_ADDR_WIDTH-1:0],virt_addr_in[VA48_INDEX_WIDTH-1:0]};
	  phys_addr_out_vld=1'b1;
	end
	if ( DO_IDLE ) begin
	  NEXT_STATE = IDLE;
	end
  end // WRITE_TLB
  DONE : begin
	if ( !DO_IDLE ) begin
	  NEXT_STATE = DONE;
	end
	if ( DO_IDLE ) begin
	  NEXT_STATE = IDLE;
	end
  end // DONE
  endcase
end // always_comb


  //
 // memory functions
//---------------------------
logic VA48_CAM_rd;
logic [VA48_CAM_ENTRY_WIDTH-1:0] VA48_CAM_rd_data;

function automatic [VA48_CAM_ENTRY_WIDTH-1:0] POP_VA48_CAM( read, [VA48_CAM_ENTRY_WIDTH-1:0] read_data );
   VA48_CAM_rd = read;
   VA48_CAM_wr_data = read_data;
   return VA48_CAM_rd_data;
endfunction

logic VA48_CAM_wr;
logic [VA48_CAM_ENTRY_WIDTH-1:0] VA48_CAM_wr_data;

function automatic void PUSH_VA48_CAM( write, [VA48_CAM_ENTRY_WIDTH-1:0] write_data );
   VA48_CAM_wr = write;
   VA48_CAM_wr_data = write_data;
endfunction

function automatic void NOP_VA48_CAM();
   VA48_CAM_rd = 1'b0;
   VA48_CAM_wr = 1'b0;
   VA48_CAM_wr_data = '0;
   VA48_CAM_rd_data = '0;
endfunction

endmodule


  //
 // memory
//---------------------------
module CAM ( clk, res_n, wr, wr_data, rd, rd_data, hit );

parameter DATA_WIDTH = 32;
parameter NUM_ENTRIES = 16;
parameter TAG_WIDTH = 8;

input  logic clk, res_n, rd, wr;
input  logic [DATA_WIDTH-1:0] wr_data;
output logic [DATA_WIDTH-1:0] rd_data;
output logic hit;
    
localparam integer ADDR_WIDTH = $clog2(NUM_ENTRIES);
localparam integer ENTRY_WIDTH = DATA_WIDTH - TAG_WIDTH;

localparam [ADDR_WIDTH-1:0] ZERO_ADDR = {ADDR_WIDTH{1'b0}};
localparam [DATA_WIDTH-1:0] ZERO_DATA = {DATA_WIDTH{1'b0}};

logic [DATA_WIDTH-1:0] MEMORY [NUM_ENTRIES-1:0];
logic [ADDR_WIDTH-1:0] HEAD, HEAD_PLUS_1; 
logic [NUM_ENTRIES-1:0] compare;

initial begin
    for ( int i = 0; i < NUM_ENTRIES; i = i + 1 ) begin
        MEMORY[ i ] = ZERO_DATA;
    end
    HEAD = ZERO_ADDR;
end

// content addressable compare function
assign compare = ( rd && !wr && res_n ) ? compare_fcn() : '0;

// in-cycle read access
assign rd_data = ( rd && !wr && res_n ) ? read_fcn() : ZERO_DATA;

// append entries (wrap)
assign HEAD_PLUS_1 = HEAD + 1'b1;

always @ ( posedge clk ) begin

  if( res_n ) begin
    if( wr && !rd ) begin // WRITE ACCESS
      MEMORY[HEAD] <= wr_data;
      HEAD <= HEAD_PLUS_1;
    end
  end else begin
      HEAD <= clear_fcn();
  end
end

function automatic [NUM_ENTRIES-1:0] compare_fcn();
  for( int i=0; i<NUM_ENTRIES; i++ ) begin
    compare_fcn[i] = ( MEMORY[i][DATA_WIDTH-1:ENTRY_WIDTH] == wr_data[DATA_WIDTH-1:ENTRY_WIDTH] );
  end
endfunction

function automatic [DATA_WIDTH-1:0] read_fcn();
  read_fcn = ZERO_DATA;
  hit = 1'b0;
  for( int i=0; i<NUM_ENTRIES; i++ )
    if( 1'b1 == compare[i] ) begin
      read_fcn = MEMORY[i];
      hit = rd && !wr && compare[i];
      break;	
    end
endfunction

function automatic [ADDR_WIDTH-1:0] clear_fcn();
  for( int i=0; i<NUM_ENTRIES; i++ )
      MEMORY[ i ] <= ZERO_DATA;
  clear_fcn = ZERO_ADDR;
endfunction

endmodule
