#include <iostream>
#include <verilated.h>
#include "VTWE.h"

using namespace std;

// Current simulation time (64-bit unsigned)
vluint64_t main_time = 0;
vluint64_t reset_time = 5;
vluint64_t unlock_time = 10;
vluint64_t start_time = 12;
vluint64_t request_time_1 = 20;
vluint64_t request_time_2 = 150;
vluint64_t max_time = 200;

vluint64_t mem_delay = 20;
vluint64_t mem_access_start = 0;
vluint64_t level = 0;
vluint64_t levels = 4;

vluint64_t twe_active = 0;
vluint64_t twe_finished = 0;
vluint64_t twe_finished_time = 0;

vluint64_t mem_access_active = 0;

vluint64_t finish_delay = 10;

// Called by $time in Verilog
double sc_time_stamp() {
    return main_time;  // Note does conversion to real, to match SystemC
}

int main(int argc, char** argv, char** env){
  
  // Prevent unused variable warnings
  if (false && argc && argv && env) {}

  // Set debug level, 0 is off, 9 is highest presently used
  // May be overridden by commandArgs
  Verilated::debug(0);

  // Randomization reset policy
  // May be overridden by commandArgs
  Verilated::randReset(2);

  // Verilator must compute traced signals
  Verilated::traceEverOn(true);

  // Pass arguments so Verilated code can see them, e.g. $value$plusargs
  // This needs to be called before you create any model
  Verilated::commandArgs(argc, argv);

  // Create logs/ directory in case we have traces to put under it
  Verilated::mkdir("logs");

  // Construct the Verilated model, from Vtop.h generated from Verilating "top.v"
  VTWE* top = new VTWE();

  top->clk = 0;
  top->res_n = !0;

  top->lock = 0;
  top->unlock = 0;
  top->key = 0x0000;
  
  top->virt_addr_in_vld = 0;
  top->virt_addr_in = 0x000000000000; // 48 bit

  top->data_from_memory_vld = 0;
  top->data_from_memory = 0x0000000000; // 40 bit
    
  top->pte_root_addr = 0x000000000000; // 48 bit

  top->mode = 0b00;

  while (!Verilated::gotFinish()) {

    main_time++;

    // toggle clock
    top->clk=!top->clk;

    if(0==top->clk){
      
      if ( main_time <= reset_time ) {
	top->res_n = !1;
      } else {
	top->res_n = !0;
      }

      if( main_time>reset_time && main_time<=unlock_time ) {
	top->key = 0xAFFA;
	top->unlock = 1;
      } else {
	top->key = 0x0000;
	top->unlock = 0;
      }

      if (main_time>=start_time && main_time<=max_time) {
	
	if(0==twe_active){
	  if((main_time==request_time_1)){
	    // request address translation 1
	    top->pte_root_addr = 0xFFFFFFFFF000;
	    top->virt_addr_in  = 0x123456789AFF;
	    top->virt_addr_in_vld = 1;
	    twe_active = 1;
	  }	
	  if((0==twe_active) && (main_time==request_time_2)){
	    // request address translation 2
	    top->pte_root_addr = 0xFFFFFFFFF000;
	    top->virt_addr_in  = 0x123456789A00;
	    top->virt_addr_in_vld = 1;
	    twe_active = 1;
	  }
	}
	
	else
	  
	  {
	    // memory access for table walk
	    if((1==top->data_to_memory_vld) && (0==mem_access_active)){
	      mem_access_active = 1;
	      mem_access_start = main_time;
	    }	
	    // page return from table walk
	    if((1==mem_access_active) && (main_time==(mem_access_start+mem_delay))){
	      top->data_from_memory_vld = 1;
	      if( 0==level ) top->data_from_memory = 0xA000000000;
	      if( 1==level ) top->data_from_memory = 0xB111111111;
	      if( 2==level ) top->data_from_memory = 0xC222222222;
	      if( 3==level ) top->data_from_memory = 0xD333333333;	  
	      mem_access_active = 0;
	    }
	    if((0==mem_access_active) && (main_time==(mem_access_start+mem_delay+2))){
	      level = (level+1)%levels;
	      top->data_from_memory_vld = 0;
	    }
	    // check for finish
	    if((1==top->twe_active) && (1==top->phys_addr_out_vld)){	    
	      twe_finished = 1;
	      twe_finished_time = main_time;
	    }
	    if((1==twe_finished) && (main_time==(twe_finished_time+finish_delay))){
	      top->virt_addr_in  = 0x000000000000;
	      top->virt_addr_in_vld = 0;	    
	      top->data_from_memory_vld = 0;
	      top->data_from_memory = 0x0000000000;
	      twe_finished = 0;
	      twe_active = 0;
	    }
	  }

      } // end if (main_time>=start_time && main_time<=max_time)
      
      if (main_time>=max_time) {	
	top->lock=1; top->unlock=1; top->res_n=0;
      }
	
    } // end if(!top->clk)
       
    // Evaluate model
    // (If you have multiple models being simulated in the same
    // timestep then instead of eval(), call eval_step() on each, then
    // eval_end_step() on each.)
    top->eval();
    
    if(!top->clk){
      VL_PRINTF( "[%3" VL_PRI64 "d] clk=%x res_n=%x lock=%x ( unlock=%x key=%4x ) PTE_ROOT=%12lx ( VA_in_vld=%x VA_in=%12lx ) ( FROM_MEM_vld=%x FROM_MEM=%12lx ) => ( TO_MEM_vld=%x TO_MEM=%12lx ) ( PA_vld=%x PA=%12lx ) locked=%x \n",
		 main_time,
		 top->clk, top->res_n,
		 top->lock, top->unlock, top->key,
		 top->pte_root_addr,
		 top->virt_addr_in_vld, top->virt_addr_in, 
		 top->data_from_memory_vld, top->data_from_memory, 
		 top->data_to_memory_vld, top->data_to_memory, 
		 top->phys_addr_out_vld, top->phys_addr_out,
		 top->locked );         
      }
  }
  
  top->final();

  //  Coverage analysis (since test passed)
#if VM_COVERAGE
  Verilated::mkdir("logs");
  VerilatedCov::write("logs/coverage.dat");
#endif

  // Delete model
  delete top;
  top = nullptr;

  // Exit
  exit(0);
  
}
